# coding: utf-8
# pylint: disable=too-many-lines,too-many-public-methods
from time import sleep
from datetime import datetime
from operator import attrgetter
from functools import lru_cache, partialmethod
from itertools import chain

from first import first

import ujson as json

from . import logger
from .cache import Playlist, AudioFeatures, db, select, db_session
from .mixins import AuthMixin, EmailMixin
from .result import SpotifyResult
from .constants import API, DEVICE_ID_RE, PLAYLIST_URI_RE, TimeRange, AudioFeature
from .exceptions import (
    SpotifyException,
    SpotifyAuthException,
    SpotifyForbiddenException,
    SpotifyRateLimitException,
)


class SpotifyClient(AuthMixin, EmailMixin):

    def __init__(self, *args, proxies=None, requests_timeout=None, **kwargs):
        '''
        Create a Spotify API object.

        :param proxies: Definition of proxies
        :param requests_timeout: Tell Requests to stop waiting for a response after a given number of seconds
        '''
        super().__init__(*args, **kwargs)
        self.proxies = proxies
        self.requests_timeout = requests_timeout

    @db_session
    def _increment_api_call_count(self):
        try:
            user = self.user
        except Exception:
            logger.warning(f'Tried to use an inexistent user: {self.user_id}')
        else:
            user.api_calls += 1
            user.last_usage_at = datetime.utcnow()

    @staticmethod
    def _check_response(response):
        try:
            response.raise_for_status()
        except:
            exception_params = {
                'status_code': response.status_code,
                'url': response.url,
                'headers': response.headers,
                'text': response.text,
            }
            if response.status_code == 429 or (
                response.status_code >= 500 and response.status_code < 600
            ):
                raise SpotifyRateLimitException(
                    retry_after=int(response.headers.get('Retry-After', 0)),
                    **exception_params,
                )

            elif response.status_code == 403:
                raise SpotifyForbiddenException(**exception_params)

            else:
                raise SpotifyException(**exception_params)

    def _internal_call(self, method, url, payload, params, headers=None):
        logger.debug(url)
        if not isinstance(payload, (bytes, str)):
            payload = json.dumps(payload)
        r = self.session.request(
            method,
            url,
            proxies=self.proxies,
            timeout=self.requests_timeout,
            headers={'Content-Type': 'application/json', **(headers or {})},
            data=payload,
            params={k: v for k, v in params.items() if v is not None},
            client_id=self.client_id,
            client_secret=self.client_secret,
        )
        logger.debug('HTTP Status Code: {r.status_code}')
        logger.debug(f'{method}: {r.url}')
        if payload and not isinstance(payload, bytes):
            logger.debug(f'DATA: {payload}')
        try:
            self._check_response(r)
        except SpotifyRateLimitException as exc:
            logger.warning(
                f'Reached API rate limit. Retrying in {exc.retry_after} seconds...'
            )
            sleep(exc.retry_after)
            return self._internal_call(method, url, payload, params)

        if self.user_id:
            self._increment_api_call_count()
        if r.text and r.text != 'null':
            results = r.json()
            logger.debug(f'RESP: {r.text}')
            return SpotifyResult(results, _client=self)

        return None

    def _api_call(self, method, url, args=None, payload=None, headers=None, **kwargs):
        if not self.is_authenticated:
            raise SpotifyAuthException

        if args:
            kwargs.update(args)
        if 'device_id' in kwargs:
            kwargs['device_id'] = self.get_device_id(kwargs['device_id'])
        if not url.startswith('http'):
            url = API.PREFIX.value + url
        return self._internal_call(method, url, payload, kwargs, headers)

    def _get(self, url, args=None, payload=None, **kwargs):
        return self._api_call('GET', url, args, payload, **kwargs)

    def _post(self, url, args=None, payload=None, **kwargs):
        return self._api_call('POST', url, args, payload, **kwargs)

    def _delete(self, url, args=None, payload=None, **kwargs):
        return self._api_call('DELETE', url, args, payload, **kwargs)

    def _put(self, url, args=None, payload=None, **kwargs):
        return self._api_call('PUT', url, args, payload, **kwargs)

    def previous(self, result):
        ''' returns the previous result given a paged result

            Parameters:
                - result - a previously returned paged result
        '''
        if result['previous']:
            return self._get(result['previous'])

        return None

    def track(self, track_id):
        ''' returns a single track given the track's ID, URI or URL

            Parameters:
                - track_id - a spotify URI, URL or ID
        '''
        _id = self._get_track_id(track_id)
        return self._get(API.TRACK.value.format(id=_id))  # pylint: disable=no-member

    def tracks(self, tracks, market='from_token'):
        ''' returns a list of tracks given a list of track IDs, URIs, or URLs

            Parameters:
                - tracks - a list of spotify URIs, URLs or IDs
                - market - an ISO 3166-1 alpha-2 country code.
        '''
        track_list = map(self._get_track_id, tracks)
        return self._get(API.TRACKS.value, ids=','.join(track_list), market=market)

    def artist(self, artist_id):
        ''' returns a single artist given the artist's ID, URI or URL

            Parameters:
                - artist_id - an artist ID, URI or URL
        '''
        _id = self._get_artist_id(artist_id)
        return self._get(API.ARTIST.value.format(id=_id))  # pylint: disable=no-member

    def artists(self, artists):
        ''' returns a list of artists given the artist IDs, URIs, or URLs

            Parameters:
                - artists - a list of  artist IDs, URIs or URLs
        '''
        artist_list = map(self._get_artist_id, artists)
        return self._get(API.ARTISTS.value, ids=','.join(artist_list))

    def artist_albums(
        self, artist_id, album_type=None, country=None, limit=20, offset=0
    ):
        ''' Get Spotify catalog information about an artist's albums

            Parameters:
                - artist_id - the artist ID, URI or URL
                - album_type - 'album', 'single', 'appears_on', 'compilation'
                - country - limit the response to one particular country.
                - limit  - the number of albums to return
                - offset - the index of the first album to return
        '''
        _id = self._get_artist_id(artist_id)
        return self._get(
            API.ARTIST_ALBUMS.value.format(id=_id),  # pylint: disable=no-member
            album_type=album_type,
            country=country,
            limit=limit,
            offset=offset,
        )

    def artist_top_tracks(self, artist_id, country='US'):
        ''' Get Spotify catalog information about an artist's top 10 tracks
            by country.

            Parameters:
                - artist_id - the artist ID, URI or URL
                - country - limit the response to one particular country.
        '''
        _id = self._get_artist_id(artist_id)
        # pylint: disable=no-member
        return self._get(API.ARTIST_TOP_TRACKS.value.format(id=_id), country=country)

    @lru_cache(maxsize=128)
    def artist_related_artists(self, artist_id):
        ''' Get Spotify catalog information about artists similar to an
            identified artist. Similarity is based on analysis of the
            Spotify community's listening history.

            Parameters:
                - artist_id - the artist ID, URI or URL
        '''
        _id = self._get_artist_id(artist_id)
        # pylint: disable=no-member
        return self._get(API.ARTIST_RELATED_ARTISTS.value.format(id=_id))

    def album(self, album_id):
        ''' returns a single album given the album's ID, URIs or URL

            Parameters:
                - album_id - the album ID, URI or URL
        '''
        _id = self._get_album_id(album_id)
        return self._get(API.ALBUM.value.format(id=_id))  # pylint: disable=no-member

    def album_tracks(self, album_id, limit=50, offset=0):
        ''' Get Spotify catalog information about an album's tracks

            Parameters:
                - album_id - the album ID, URI or URL
                - limit  - the number of items to return
                - offset - the index of the first item to return
        '''
        _id = self._get_album_id(album_id)
        # pylint: disable=no-member
        return self._get(
            API.ALBUM_TRACKS.value.format(id=_id), limit=limit, offset=offset
        )

    def albums(self, albums):
        ''' returns a list of albums given the album IDs, URIs, or URLs

            Parameters:
                - albums - a list of  album IDs, URIs or URLs
        '''
        album_list = map(self._get_album_id, albums)
        return self._get(API.ALBUMS.value, ids=','.join(album_list))

    def search(self, url, q, limit=10, offset=0, market='from_token'):
        ''' searches for an item

            Parameters:
                - q - the search query
                - limit  - the number of items to return
                - offset - the index of the first item to return
                - type - the type of item to return. One of 'artist', 'album',
                         'track' or 'playlist'
                - market - An ISO 3166-1 alpha-2 country code or the string from_token.
        '''
        return self._get(url, q=q, limit=limit, offset=offset, market=market)

    def search_track(self, track, limit=10, offset=0, market='from_token'):
        return self.search(
            API.SEARCH_TRACK.value, track, limit=limit, offset=offset, market=market
        )

    def search_album(self, album, limit=10, offset=0, market='from_token'):
        return self.search(
            API.SEARCH_ALBUM.value, album, limit=limit, offset=offset, market=market
        )

    def search_artist(self, artist, limit=10, offset=0, market='from_token'):
        return self.search(
            API.SEARCH_ARTIST.value, artist, limit=limit, offset=offset, market=market
        )

    def search_playlist(self, playlist, limit=10, offset=0, market='from_token'):
        return self.search(
            API.SEARCH_PLAYLIST.value,
            playlist,
            limit=limit,
            offset=offset,
            market=market,
        )

    def profile(self, user):
        ''' Gets basic profile information about a Spotify User

            Parameters:
                - user - the id of the user
        '''
        # pylint: disable=no-member
        return self._get(API.USER.value.format(user_id=user))

    def current_user_playlists(self, limit=50, offset=0):
        """ Get current user playlists without required getting his profile
            Parameters:
                - limit  - the number of items to return
                - offset - the index of the first item to return
        """
        return self._get(API.MY_PLAYLISTS.value, limit=limit, offset=offset)

    def user_playlists(self, user, limit=50, offset=0):
        ''' Gets playlists of a user

            Parameters:
                - user - the id of the usr
                - limit  - the number of items to return
                - offset - the index of the first item to return
        '''
        # pylint: disable=no-member
        return self._get(
            API.PLAYLISTS.value.format(user_id=user), limit=limit, offset=offset
        )

    def user_playlist(self, user, playlist_id=None, fields=None):
        ''' Gets playlist of a user
            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - fields - which fields to return
        '''
        if playlist_id is None:
            return self._get("users/%s/starred" % (user), fields=fields)

        _id = self._get_playlist_id(playlist_id)
        # pylint: disable=no-member
        return self._get(
            API.PLAYLIST.value.format(user_id=user, playlist_id=_id), fields=fields
        )

    def user_playlist_tracks(
        self,
        user,
        playlist_id=None,
        fields=None,
        limit=100,
        offset=0,
        market='from_token',
    ):
        ''' Get full details of the tracks of a playlist owned by a user.

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - fields - which fields to return
                - limit - the maximum number of tracks to return
                - offset - the index of the first track to return
                - market - an ISO 3166-1 alpha-2 country code.
        '''
        _id = self._get_playlist_id(playlist_id)
        # pylint: disable=no-member
        return self._get(
            API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id),
            limit=limit,
            offset=offset,
            fields=fields,
            market=market,
        )

    def user_playlist_create(self, user, name, public=True, description=''):
        ''' Creates a playlist for a user

            Parameters:
                - user - the id of the user
                - name - the name of the playlist
                - public - is the created playlist public
                - description - the description of the playlist
        '''
        # pylint: disable=no-member
        data = {'name': name, 'public': public, 'description': description}
        return self._post(API.PLAYLISTS.value.format(user_id=user), payload=data)

    def user_playlist_upload_cover_image(self, user, playlist_id, image):
        ''' Creates a playlist for a user

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - image - base64 encoded image
        '''
        # pylint: disable=no-member
        return self._put(
            API.PLAYLIST_IMAGES.value.format(user_id=user, playlist_id=playlist_id),
            payload=image,
            headers={'Content-Type': 'image/jpeg'},
        )

    def user_playlist_change_details(
        self,
        user,
        playlist_id,
        name=None,
        public=None,
        collaborative=None,
        description=None,
    ):
        ''' Changes a playlist's name and/or public/private state

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - name - optional name of the playlist
                - public - optional is the playlist public
                - collaborative - optional is the playlist collaborative
                - description - the description of the playlist
        '''
        data = {}
        if isinstance(name, str):
            data['name'] = name
        if isinstance(public, bool):
            data['public'] = public
        if isinstance(collaborative, bool):
            data['collaborative'] = collaborative
        if isinstance(collaborative, str):
            data['description'] = description
        # pylint: disable=no-member
        return self._put(
            API.PLAYLIST.value.format(user_id=user, playlist_id=playlist_id),
            payload=data,
        )

    def user_playlist_unfollow(self, user, playlist_id):
        ''' Unfollows (deletes) a playlist for a user

            Parameters:
                - user - the id of the user
                - name - the name of the playlist
        '''
        return self._delete("users/%s/playlists/%s/followers" % (user, playlist_id))

    def user_playlist_add_tracks(self, user, playlist_id, tracks, position=None):
        ''' Adds tracks to a playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - tracks - a list of track URIs, URLs or IDs
                - position - the position to add the tracks
        '''
        _id = self._get_playlist_id(playlist_id)
        # pylint: disable=no-member
        url = API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id)
        track_uris = list(map(self._get_track_uri, tracks))
        if len(track_uris) <= 100:
            return self._post(url, payload={"uris": track_uris}, position=position)

        batches = [
            {"uris": track_uris[i:i + 100]} for i in range(0, len(track_uris), 100)
        ]
        results = [
            self._post(
                url,
                payload=t,
                position=(None if position is None else i * 100 + position),
            )
            for i, t in enumerate(batches)
        ]
        return results

    def user_playlist_replace_tracks(self, user, playlist_id, tracks):
        ''' Replace all tracks in a playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - tracks - the list of track ids to add to the playlist
        '''
        _id = self._get_playlist_id(playlist_id)
        # pylint: disable=no-member
        url = API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id)
        first_100_tracks, rest_tracks = tracks[:100], tracks[100:]
        track_uris = list(map(self._get_track_uri, first_100_tracks))
        replaced = self._put(url, payload={"uris": track_uris})
        if not rest_tracks:
            return replaced

        added = self.user_playlist_add_tracks(user, playlist_id, rest_tracks)
        if isinstance(added, list):
            return [replaced, *added]

        return [replaced, added]

    def user_playlist_reorder_tracks(
        self,
        user,
        playlist_id,
        range_start,
        insert_before,
        range_length=1,
        snapshot_id=None,
    ):
        ''' Reorder tracks in a playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - range_start - the position of the first track to be reordered
                - range_length - optional the number of tracks to be reordered (default: 1)
                - insert_before - the position where the tracks should be inserted
                - snapshot_id - optional playlist's snapshot ID
        '''
        _id = self._get_playlist_id(playlist_id)
        payload = {
            "range_start": range_start,
            "range_length": range_length,
            "insert_before": insert_before,
        }
        if snapshot_id:
            payload["snapshot_id"] = snapshot_id
        # pylint: disable=no-member
        return self._put(
            API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id),
            payload=payload,
        )

    def user_playlist_remove_all_occurrences_of_tracks(
        self, user, playlist_id, tracks, snapshot_id=None
    ):
        ''' Removes all occurrences of the given tracks from the given playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - tracks - the list of track ids to add to the playlist
                - snapshot_id - optional id of the playlist snapshot

        '''
        _id = self._get_playlist_id(playlist_id)
        track_uris = map(self._get_track_uri, tracks)
        payload = {"tracks": [{"uri": track} for track in track_uris]}
        if snapshot_id:
            payload["snapshot_id"] = snapshot_id
        # pylint: disable=no-member
        return self._delete(
            API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id),
            payload=payload,
        )

    def user_playlist_remove_specific_occurrences_of_tracks(
        self, user, playlist_id, tracks, snapshot_id=None
    ):
        ''' Removes all occurrences of the given tracks from the given playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - tracks - an array of objects containing Spotify URIs of the tracks
                           to remove with their current positions in the playlist.
                    For example:
                        [
                          { "uri":"4iV5W9uYEdYUVa79Axb7Rh", "positions":[2] },
                          { "uri":"1301WleyT98MSxVHPZCA6M", "positions":[7] }
                        ]
                - snapshot_id - optional id of the playlist snapshot
        '''
        _id = self._get_playlist_id(playlist_id)
        ftracks = []
        for tr in tracks:
            ftracks.append(
                {"uri": self._get_uri("track", tr["uri"]), "positions": tr["positions"]}
            )
        payload = {"tracks": ftracks}
        if snapshot_id:
            payload["snapshot_id"] = snapshot_id
        # pylint: disable=no-member
        return self._delete(
            API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id),
            payload=payload,
        )

    def user_playlist_follow_playlist(self, playlist_owner_id, playlist_id):
        '''
        Add the current authenticated user as a follower of a playlist.

        Parameters:
            - playlist_owner_id - the user id of the playlist owner
            - playlist_id - the id of the playlist

        '''
        return self._put(
            API.PLAYLIST_FOLLOWERS.value.format(  # pylint: disable=no-member
                owner_id=playlist_owner_id, playlist_id=playlist_id
            )
        )

    def user_playlist_is_following(self, playlist_owner_id, playlist_id, user_ids):
        '''
        Check to see if the given users are following the given playlist

        Parameters:
            - playlist_owner_id - the user id of the playlist owner
            - playlist_id - the id of the playlist
            - user_ids - the ids of the users that you want to check to see if they follow the playlist. Maximum: 5 ids.

        '''
        return self._get(
            API.PLAYLIST_FOLLOWERS_CONTAINS.value.format(  # pylint: disable=no-member
                user_id=playlist_owner_id, playlist_id=playlist_id
            ),
            ids=','.join(user_ids),
        )

    def me(self):
        ''' Get detailed profile information about the current user.
            An alias for the 'current_user' method.
        '''
        return self._get(API.ME.value)

    def current_user(self):
        ''' Get detailed profile information about the current user.
            An alias for the 'me' method.
        '''
        return self.me()

    def current_user_saved_albums(self, limit=20, offset=0):
        ''' Gets a list of the albums saved in the current authorized user's
            "Your Music" library

            Parameters:
                - limit - the number of albums to return
                - offset - the index of the first album to return

        '''
        return self._get(API.MY_ALBUMS.value, limit=limit, offset=offset)

    def current_user_saved_tracks(self, limit=20, offset=0):
        ''' Gets a list of the tracks saved in the current authorized user's
            "Your Music" library

            Parameters:
                - limit - the number of tracks to return
                - offset - the index of the first track to return

        '''
        return self._get(API.MY_TRACKS.value, limit=limit, offset=offset)

    def current_user_followed_artists(self, limit=20, after=None):
        ''' Gets a list of the artists followed by the current authorized user

            Parameters:
                - limit - the number of tracks to return
                - after - ghe last artist ID retrieved from the previous request

        '''
        return self._get(
            API.MY_FOLLOWING.value, type='artist', limit=limit, after=after
        )

    def user_follow_artists(self, ids=None):
        ''' Follow one or more artists
            Parameters:
                - ids - a list of artist IDs
        '''
        return self._put(API.MY_FOLLOWING.value, type='artist', ids=','.join(ids or []))

    def user_follow_users(self, ids=None):
        ''' Follow one or more users
            Parameters:
                - ids - a list of user IDs
        '''
        return self._put(API.MY_FOLLOWING.value, type='user', ids=','.join(ids or []))

    def current_user_saved_tracks_delete(self, tracks=None):
        ''' Remove one or more tracks from the current user's
            "Your Music" library.

            Parameters:
                - tracks - a list of track URIs, URLs or IDs
        '''
        track_list = []
        if tracks is not None:
            track_list = map(self._get_track_id, tracks)
        return self._delete(API.MY_TRACKS.value, ids=','.join(track_list))

    def current_user_saved_tracks_contains(self, tracks=None):
        ''' Check if one or more tracks is already saved in
            the current Spotify user’s “Your Music” library.

            Parameters:
                - tracks - a list of track URIs, URLs or IDs
        '''
        track_list = []
        if tracks is not None:
            track_list = map(self._get_track_id, tracks)
        return self._get(API.MY_TRACKS_CONTAINS.value, ','.join(track_list))

    def current_user_saved_tracks_add(self, tracks=None):
        ''' Add one or more tracks to the current user's
            "Your Music" library.

            Parameters:
                - tracks - a list of track URIs, URLs or IDs
        '''
        track_list = []
        if tracks is not None:
            track_list = map(self._get_track_id, tracks)
        return self._put(API.MY_TRACKS.value, ids=','.join(track_list))

    def current_user_top_artists(
        self, limit=20, offset=0, time_range=TimeRange.MEDIUM_TERM
    ):
        ''' Get the current user's top artists

            Parameters:
                - limit - the number of entities to return
                - offset - the index of the first entity to return
                - time_range - Over what time frame are the affinities computed
                  Valid-values: short_term, medium_term, long_term
        '''
        return self._get(
            API.MY_TOP.value.format(type='artists'),  # pylint: disable=no-member
            time_range=TimeRange(time_range).value,
            limit=limit,
            offset=offset,
        )

    def current_user_top_tracks(
        self, limit=20, offset=0, time_range=TimeRange.MEDIUM_TERM
    ):
        ''' Get the current user's top tracks

            Parameters:
                - limit - the number of entities to return
                - offset - the index of the first entity to return
                - time_range - Over what time frame are the affinities computed
                  Valid-values: short_term, medium_term, long_term
        '''
        return self._get(
            API.MY_TOP.value.format(type='tracks'),  # pylint: disable=no-member
            time_range=TimeRange(time_range).value,
            limit=limit,
            offset=offset,
        )

    def current_user_saved_albums_add(self, albums=None):
        ''' Add one or more albums to the current user's
            "Your Music" library.
            Parameters:
                - albums - a list of album URIs, URLs or IDs
        '''
        album_list = map(self._get_album_id, albums or [])
        return self._put(API.MY_ALBUMS.value, ids=','.join(album_list))

    def featured_playlists(
        self, locale=None, country=None, timestamp=None, limit=20, offset=0
    ):
        ''' Get a list of Spotify featured playlists

            Parameters:
                - locale - The desired language, consisting of a lowercase ISO
                  639 language code and an uppercase ISO 3166-1 alpha-2 country
                  code, joined by an underscore.

                - country - An ISO 3166-1 alpha-2 country code.

                - timestamp - A timestamp in ISO 8601 format:
                  yyyy-MM-ddTHH:mm:ss. Use this parameter to specify the user's
                  local time to get results tailored for that specific date and
                  time in the day

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 50

                - offset - The index of the first item to return. Default: 0
                  (the first object). Use with limit to get the next set of
                  items.
        '''
        return self._get(
            API.FEATURED_PLAYLISTS.value,
            locale=locale,
            country=country,
            timestamp=timestamp,
            limit=limit,
            offset=offset,
        )

    def new_releases(self, country=None, limit=20, offset=0):
        ''' Get a list of new album releases featured in Spotify

            Parameters:
                - country - An ISO 3166-1 alpha-2 country code.

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 50

                - offset - The index of the first item to return. Default: 0
                  (the first object). Use with limit to get the next set of
                  items.
        '''
        return self._get(
            API.NEW_RELEASES.value, country=country, limit=limit, offset=offset
        )

    def categories(self, country=None, locale=None, limit=20, offset=0):
        ''' Get a list of new album releases featured in Spotify

            Parameters:
                - country - An ISO 3166-1 alpha-2 country code.
                - locale - The desired language, consisting of an ISO 639
                  language code and an ISO 3166-1 alpha-2 country code, joined
                  by an underscore.

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 50

                - offset - The index of the first item to return. Default: 0
                  (the first object). Use with limit to get the next set of
                  items.
        '''
        return self._get(
            API.CATEGORIES.value,
            country=country,
            locale=locale,
            limit=limit,
            offset=offset,
        )

    def category_playlists(self, category_id=None, country=None, limit=20, offset=0):
        ''' Get a list of new album releases featured in Spotify

            Parameters:
                - category_id - The Spotify category ID for the category.

                - country - An ISO 3166-1 alpha-2 country code.

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 50

                - offset - The index of the first item to return. Default: 0
                  (the first object). Use with limit to get the next set of
                  items.
        '''
        # pylint: disable=no-member
        return self._get(
            API.CATEGORY_PLAYLISTS.value.format(id=category_id),
            country=country,
            limit=limit,
            offset=offset,
        )

    def recommendations(
        self,
        seed_artists=None,
        seed_genres=None,
        seed_tracks=None,
        limit=20,
        country='from_token',
        **kwargs,
    ):
        ''' Get a list of recommended tracks for one to five seeds.

            Parameters:
                - seed_artists - a list of artist IDs, URIs or URLs

                - seed_tracks - a list of artist IDs, URIs or URLs

                - seed_genres - a list of genre names. Available genres for
                  recommendations can be found by calling recommendation_genre_seeds

                - country - An ISO 3166-1 alpha-2 country code. If provided, all
                  results will be playable in this country.

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 100

                - min/max/target_<attribute> - For the tuneable track attributes listed
                  in the documentation, these values provide filters and targeting on
                  results.
        '''
        params = dict(limit=limit)
        if seed_artists:
            params['seed_artists'] = ','.join(map(self._get_artist_id, seed_artists))
        if seed_genres:
            params['seed_genres'] = ','.join(seed_genres)
        if seed_tracks:
            params['seed_tracks'] = ','.join(map(self._get_track_id, seed_tracks))
        if country:
            params['market'] = country
        for attribute in list(AudioFeature):
            for prefix in ["min_", "max_", "target_"]:
                param = prefix + attribute.value
                if param in kwargs:
                    params[param] = kwargs[param]
        return self._get(API.RECOMMENDATIONS.value, **params)

    def recommendation_genre_seeds(self):
        ''' Get a list of genres available for the recommendations function.
        '''
        return self._get(API.RECOMMENDATIONS_GENRES.value)

    def audio_analysis(self, track=None):
        ''' Get audio analysis for a track based upon its Spotify ID
            Parameters:
                - track - a track URI, URL or ID
        '''
        _id = self._get_track_id(track)
        # pylint: disable=no-member
        return self._get(API.AUDIO_ANALYSIS.value.format(id=_id))

    def audio_features(self, track=None, tracks=None, with_cache=True):
        ''' Get audio features for one or multiple tracks based upon their Spotify IDs
            Parameters:
                - track - a track URI, URL or ID
                - tracks - a list of track URIs, URLs or IDs, maximum: 100 ids
        '''
        if track:
            _id = self._get_track_id(track)
            # pylint: disable=no-member
            return self._get(API.AUDIO_FEATURES_SINGLE.value.format(id=_id))

        tracks = list(map(self._get_track_id, tracks or []))
        cached_tracks = []
        if with_cache:
            with db_session:
                cached_tracks = select(a for a in AudioFeatures if a.id in tracks)[:]
                tracks = list(set(tracks) - {a.id for a in cached_tracks})
        batches = [tracks[i:i + 100] for i in range(0, len(tracks), 100)]
        audio_features = [
            self._get(API.AUDIO_FEATURES_MULTIPLE.value, ids=','.join(t))
            for t in batches
        ]
        with db_session:
            audio_features = [
                AudioFeatures.from_dict(t) for t in chain.from_iterable(audio_features)
            ] + cached_tracks
        return audio_features

    def devices(self):
        ''' Get a list of user's available devices.
        '''
        return self._get(API.DEVICES.value)

    @lru_cache(maxsize=128)
    def get_device_id(self, device=None):
        if isinstance(device, (str, bytes)) and DEVICE_ID_RE.match(device):
            return device

        return self.get_device(device).id

    def get_device(self, device=None):
        '''Get Spotify device based on name

        :param str, optional device: device name or ID
        :param str, optional field: device attribute to return

        str or dict: Spotify device
        '''
        devices = self.devices().devices
        device_names = ', '.join([d.name for d in devices])
        device_name_or_id = device
        if not device_name_or_id:
            device = first(devices, key=attrgetter('is_active'))
            if not device:
                raise ValueError(
                    f'''
        There's no active device.
        Possible devices: {device_names}'''
                )

        else:
            device = first(devices, key=lambda d: device_name_or_id in (d.name, d.id))
            if not device:
                raise ValueError(
                    f'''
        Device {device_name_or_id} doesn't exist.
        Possible devices: {device_names}'''
                )

        return device

    def current_playback(self, market='from_token'):
        ''' Get information about user's current playback.

            Parameters:
                - market - an ISO 3166-1 alpha-2 country code.
        '''
        return self._get(API.PLAYER.value, market=market)

    def current_user_recently_played(self, limit=50):
        ''' Get the current user's recently played tracks

            Parameters:
                - limit - the number of entities to return
        '''
        return self._get(API.RECENTLY_PLAYED.value, limit=limit)

    def currently_playing(self, market='from_token'):
        ''' Get user's currently playing track.

            Parameters:
                - market - an ISO 3166-1 alpha-2 country code.
        '''
        return self._get(API.CURRENTLY_PLAYING.value, market=market)

    def transfer_playback(self, device, force_play=True):
        ''' Transfer playback to another device.
            Note that the API accepts a list of device ids, but only
            actually supports one.

            Parameters:
                - device - transfer playback to this device
                - force_play - true: after transfer, play. false:
                               keep current state.
        '''
        device_id = self.get_device_id(device)
        data = {'device_ids': [device_id], 'play': force_play}
        return self._put(API.PLAYER.value, payload=data)

    def start_playback(
        self,
        device=None,
        artist=None,
        album=None,
        playlist=None,
        tracks=None,
        offset=None,
    ):
        ''' Start or resume user's playback.

            Parameters:
                - device - device target for playback
                - playlist - spotify playlist to play
                - artist - spotify artist to play
                - album - spotify album to play
                - tracks - spotify tracks to play
                - offset - offset into context by index or track
        '''
        data = {}
        if playlist:
            data['context_uri'] = self._get_playlist_uri(playlist)
        elif album:
            data['context_uri'] = self._get_album_uri(album)
        elif artist:
            data['context_uri'] = self._get_artist_uri(artist)
        elif tracks:
            data['uris'] = list(map(self._get_track_uri, tracks))
        if isinstance(offset, int):
            data['offset'] = dict(position=offset)
        elif isinstance(offset, str):
            data['offset'] = dict(uri=offset)
        return self._put(API.PLAY.value, device_id=device, payload=data)

    def pause_playback(self, device=None):
        ''' Pause user's playback.

            Parameters:
                - device - device target for playback
        '''
        return self._put(API.PAUSE.value, device_id=device)

    def next_track(self, device=None):
        ''' Skip user's playback to next track.

            Parameters:
                - device - device target for playback
        '''
        return self._post(API.NEXT.value, device_id=device)

    def previous_track(self, device=None):
        ''' Skip user's playback to previous track.

            Parameters:
                - device - device target for playback
        '''
        return self._post(API.PREVIOUS.value, device_id=device)

    def seek_track(self, position_ms, device=None):
        ''' Seek to position in current track.

            Parameters:
                - position_ms - position in milliseconds to seek to
                - device - device target for playback
        '''
        if not isinstance(position_ms, int):
            logger.warning('position_ms must be an integer')
            return None

        return self._put(API.SEEK.value, position_ms=position_ms, device_id=device)

    def repeat(self, state, device=None):
        ''' Set repeat mode for playback.

            Parameters:
                - state - `track`, `context`, or `off`
                - device - device target for playback
        '''
        if state not in ['track', 'context', 'off']:
            logger.warning('Invalid state')
            return

        self._put(API.REPEAT.value, state=state, device_id=device)

    def volume(self, volume_percent: int = None, device: str = None):
        ''' Get or set playback volume.

            Parameters:
                - volume_percent - volume between 0 and 100
                - device - device target for playback
        '''
        device = self.get_device(device)
        if volume_percent is None:
            return device.volume_percent

        assert 0 <= volume_percent <= 100
        return self._put(
            API.VOLUME.value, volume_percent=volume_percent, device_id=device.id
        )

    def shuffle(self, state, device=None):
        ''' Toggle playback shuffling.

            Parameters:
                - state - true or false
                - device - device target for playback
        '''
        if not isinstance(state, bool):
            logger.warning('State must be a boolean')
            return

        state = str(state).lower()
        self._put(API.SHUFFLE.value, state=state, device_id=device)

    @staticmethod
    def _get_id(_type, result):
        if isinstance(result, str):
            fields = result.split(':')
            if len(fields) >= 3:
                if _type != fields[-2]:
                    logger.warning(
                        'Expected id of type %s but found type %s %s',
                        _type,
                        fields[-2],
                        result,
                    )
                return fields[-1]

            fields = result.split('/')
            if len(fields) >= 3:
                itype = fields[-2]
                if _type != itype:
                    logger.warning(
                        'Expected id of type %s but found type %s %s',
                        _type,
                        itype,
                        result,
                    )
                return fields[-1]

        elif isinstance(result, (SpotifyResult, db.Entity)):
            return result.id

        elif isinstance(result, dict):
            return result['id']

        return result

    _get_track_id = partialmethod(_get_id, 'track')
    _get_artist_id = partialmethod(_get_id, 'artist')
    _get_album_id = partialmethod(_get_id, 'album')
    _get_playlist_id = partialmethod(_get_id, 'playlist')

    def _get_uri(self, _type, result):
        return 'spotify:' + _type + ":" + self._get_id(_type, result)

    def _get_playlist_uri(self, playlist, user=None):
        if isinstance(playlist, (str, bytes)) and PLAYLIST_URI_RE.match(playlist):
            return playlist

        if isinstance(playlist, Playlist):
            return playlist.uri

        try:
            if 'uri' in playlist:
                return playlist['uri']

        except:
            return f'spotify:user:{self._get_id("user", user)}:playlist:{self._get_id("playlist", playlist)}'

        return None

    _get_track_uri = partialmethod(_get_uri, 'track')
    _get_artist_uri = partialmethod(_get_uri, 'artist')
    _get_album_uri = partialmethod(_get_uri, 'album')



import os
import random
import threading
from collections import OrderedDict

from first import first
from cached_property import cached_property

from .. import config
from ..cache import Playlist, db_session
from ..volume import (
    AlsaVolumeControl,
    LinuxVolumeControl,
    SpotifyVolumeControl,
    ApplescriptVolumeControl,
)
from ..constants import ItemType, TimeRange, VolumeBackend


class PlayerMixin:

    def __init__(
        self,
        *args,
        device=None,
        alsa_device=None,
        alsa_mixer=None,
        speaker=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.device = device or config.player.device
        self.alsa_device = alsa_device or config.player.alsa.device
        self.alsa_mixer = alsa_mixer or config.player.alsa.mixer
        self.speaker = speaker or config.player.speaker

    @cached_property
    def _optimal_backend(self):
        return first(self._backends.values())

    @cached_property
    def _backends(self):
        return OrderedDict(
            {
                VolumeBackend.APPLESCRIPT: self._applescript_volume_control,
                VolumeBackend.LINUX: self._linux_volume_control,
                VolumeBackend.ALSA: self._alsa_volume_control,
                VolumeBackend.SPOTIFY: self._spotify_volume_control,
            }
        )

    @cached_property
    def _applescript_volume_control(self):
        try:
            # pylint: disable=no-member
            if os.uname().sysname != 'Darwin':
                return None

        except:
            return None

        return ApplescriptVolumeControl(self.speaker)

    @cached_property
    def _linux_volume_control(self):
        try:
            # pylint: disable=no-member
            if os.uname().sysname != 'Linux':
                return None

        except:
            return None

        return LinuxVolumeControl(
            self,
            self.alsa_mixer,
            spotify_device=self.device,
            alsa_device=self.alsa_device,
        )

    @cached_property
    def _alsa_volume_control(self):
        try:
            # pylint: disable=no-member
            if os.uname().sysname != 'Linux':
                return None

        except:
            return None

        return AlsaVolumeControl(self.alsa_mixer, device=self.alsa_device)

    @cached_property
    def _spotify_volume_control(self):
        return SpotifyVolumeControl(self, device=self.device)

    def backend(self, backend=None, device=None):
        if not backend:
            return self._optimal_backend

        volume_backend = self._backends[VolumeBackend(backend)]
        if not volume_backend:
            raise ValueError(
                f'Backend {volume_backend} is not available on this system'
            )

        if isinstance(
            volume_backend, SpotifyVolumeControl
        ) and device and device != self.device:
            volume_backend = SpotifyVolumeControl(self, device=device)
        return volume_backend

    def change_volume(self, value=0, backend=None, device=None):
        volume_backend = self.backend(backend, device=device)
        volume = volume_backend.volume + value
        volume_backend.volume = volume
        return volume

    def volume_up(self, backend=None):
        return self.change_volume(value=+1, backend=backend)

    def volume_down(self, backend=None):
        return self.change_volume(value=-1, backend=backend)

    def fade_up(self, **kwargs):
        self.fade(**{**config.volume.fade.up, **kwargs})

    def fade_down(self, **kwargs):
        self.fade(**{**config.volume.fade.down, **kwargs})

    #  pylint: disable=too-many-arguments

    def fade(
        self,
        limit=50,
        start=0,
        step=1,
        seconds=300,
        force=False,
        backend=None,
        spotify_volume=100,
        device=None,
    ):
        volume_backend = self.backend(backend, device=device)
        if not isinstance(volume_backend, SpotifyVolumeControl):
            self.change_volume(
                spotify_volume, backend=VolumeBackend.SPOTIFY, device=device
            )
        self.change_volume(start, backend=backend, device=device)
        kwargs = dict(
            limit=int(limit),
            start=int(start),
            step=int(step),
            seconds=int(seconds),
            force=bool(force),
        )
        threading.Thread(target=volume_backend.fade, kwargs=kwargs).start()

    @db_session
    def play_recommended_tracks(
        self, time_range=TimeRange.LONG_TERM, device=None, **kwargs
    ):
        fade_args = kwargs.get('fade_args') or {
            k[5:]: v for k, v in kwargs.items() if k.startswith('fade_')
        }
        recommendation_args = kwargs.get('recommendation_args') or {
            k[4:]: v for k, v in kwargs.items() if k.startswith('rec_')
        }
        recommendation_args['time_range'] = time_range
        tracks = self.recommend_by_top_artists(**recommendation_args)
        self.fade_up(device=device, **fade_args)
        result = self.start_playback(tracks=tracks, device=device)
        return {'playing': True, 'device': device, 'tracks': tracks, 'result': result}

    @db_session
    def play_recommended_genre(
        self, time_range=TimeRange.LONG_TERM, device=None, **kwargs
    ):
        fade_args = kwargs.get('fade_args') or {
            k[5:]: v for k, v in kwargs.items() if k.startswith('fade_')
        }
        popularity = random.choice(list(Playlist.Popularity)[:3])
        genre = self.top_genres(
            time_range=time_range
        ).select().without_distinct().random(
            1
        )[
            0
        ]
        playlist = self.genre_playlist(genre.name, popularity)
        while not playlist:
            playlist = self.genre_playlist(genre.name, popularity)
        self.fade_up(device=device, **fade_args)
        result = playlist.play(self, device=device)
        return {
            'playing': True,
            'device': device,
            'playlist': playlist.to_dict(),
            'result': result,
        }

    @db_session
    def play(
        self, time_range=TimeRange.LONG_TERM, device=None, item_type=None, **kwargs
    ):
        item_type = item_type or random.choice([ItemType.TRACKS, ItemType.PLAYLIST])
        if item_type == ItemType.TRACKS:
            return self.play_recommended_tracks(time_range, device, **kwargs)

        if item_type == ItemType.PLAYLIST:
            return self.play_recommended_genre(time_range, device, **kwargs)

        return {'playing': False}



import uuid
import socket
import asyncio
import threading
from pathlib import Path

import aiohttp
import aiohttp.web
from pony.orm import get
from oauthlib.oauth2 import BackendApplicationClient
from aiohttp.web_runner import GracefulExit

from ... import root, config, logger
from ...cache import User, select, db_session
from ...constants import API, AuthFlow, AllScopes
from ...exceptions import SpotifyCredentialsException
from .aiohttp_oauthlib import OAuth2Session

AUTH_HTML_FILE = root / 'html' / 'auth_message.html'
CACHE_FILE = Path.home() / '.cache' / 'spfy' / '.web_cache'
web_app = aiohttp.web.Application()


def run_app():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    aiohttp.web.run_app(
        web_app, host='0.0.0.0', port=config.auth.callback.port, handle_signals=False
    )


class AuthMixin:

    def __init__(
        self,
        *args,
        client_id=None,
        client_secret=None,
        redirect_uri=None,
        user_id=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.client_id = client_id or config.app.client_id
        self.client_secret = client_secret or config.app.client_secret
        self.redirect_uri = self._get_redirect_uri(redirect_uri)
        self.session = None
        self.user_id = user_id
        self.callback_reached = threading.Event()

    @staticmethod
    def _get_redirect_uri(redirect_uri):
        redirect_uri = (
            redirect_uri
            or config.app.redirect_uri
            or f'http://{socket.gethostname()}.local'
        )
        if config.auth.callback.enabled and config.auth.callback.port and redirect_uri:
            redirect_uri += f':{config.auth.callback.port}'
        return redirect_uri

    @property
    @db_session
    def user(self):
        return User[self.user_id]

    @property
    def is_authenticated(self):
        return bool(self.session and self.session.authorized)

    async def authenticate_user(
        self,
        username=None,
        email=None,
        code=None,
        state=None,
        auth_response=None,
        scope=AllScopes,
    ):
        session = self.session or OAuth2Session(
            self.client_id,
            redirect_uri=self.redirect_uri,
            scope=scope,
            auto_refresh_url=API.TOKEN.value,
        )
        with db_session:
            if self.user_id:
                user = User.get(id=self.user_id)
                if user and user.token:
                    session.token = user.token
                    session.token_updater = User.token_updater(user.id)
                    return session

            if username or email:
                if username:
                    user = get(u for u in User if u.username == username)
                elif email:
                    user = get(u for u in User if u.email == email)
                if user:
                    self.user_id = user.id
                    session.token = user.token
                    session.token_updater = User.token_updater(user.id)
                    return session

            if code or auth_response:
                token = await session.fetch_token(
                    token_url=API.TOKEN.value,
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                    code=code,
                    state=state,
                    authorization_response=auth_response,
                )
                user_details = await self.current_user()
                user = select(
                    u
                    for u in User
                    if u.username == user_details.id or u.email == user_details.email
                ).for_update().get()
                if user:
                    user.token = token
                    if user.id != self.user_id:
                        self.user_id = user.id
                elif self.user_id:
                    user = User.get_for_update(id=self.user_id)
                    if user:
                        user.token = token
                if not user:
                    self.user_id = self.user_id or uuid.uuid4()
                    user_details['user_id'] = self.user_id
                    user_details['token'] = token
                    user = User.from_dict(user_details)
                return session

        return session

    @db_session
    async def authenticate_server(self):
        default_user = User.default()
        self.user_id = default_user.id
        session = self.session or OAuth2Session(
            client=BackendApplicationClient(self.client_id)
        )
        session.token_updater = User.token_updater(default_user.id)
        if default_user.token:
            session.token = default_user.token
        else:
            default_user.token = await session.fetch_token(
                token_url=API.TOKEN.value,
                client_id=self.client_id,
                client_secret=self.client_secret,
            )
        return session

    async def authenticate(self, flow=config.auth.flow, **auth_params):
        if not (self.client_id and self.client_secret):
            raise SpotifyCredentialsException

        flow = AuthFlow(flow)
        if flow == AuthFlow.CLIENT_CREDENTIALS:
            self.session = await self.authenticate_server()
        elif flow == AuthFlow.AUTHORIZATION_CODE:
            self.session = await self.authenticate_user(**auth_params)
            if not self.session.token:
                if config.auth.callback.enabled:
                    self.start_callback()
                authorization_url, _ = self.session.authorization_url(
                    API.AUTHORIZE.value
                )
                if config.auth.send_auth_url_to_email:
                    email = auth_params.get('email') or config.auth.email
                    self.send_auth_email(email, authorization_url)
                else:
                    print(f'Login here: {authorization_url}')
                await self.wait_for_authorization()

    async def wait_for_authorization(self):
        if not config.auth.callback.enabled:
            url = input('Paste the URL you are redirected to:')
            self.session = await self.authenticate_user(auth_response=url)
        else:
            self.callback_reached.wait(config.auth.callback.timeout)
            await self.stop_callback()

    @staticmethod
    async def stop_callback():
        try:
            async with aiohttp.request(
                'GET', f'http://localhost:{config.auth.callback.port}'
            ) as resp:
                await resp.read()
        except aiohttp.client_exceptions.ServerDisconnectedError:
            pass

    def start_callback(self):
        self.callback_reached.clear()

        async def callback(request):
            if self.callback_reached.is_set():
                raise GracefulExit

            code = request.query['code']
            state = request.query.get('code')
            html = AUTH_HTML_FILE.read_text()  # pylint: disable=no-member
            try:
                self.session = asyncio.run_coroutine_threadsafe(
                    self.authenticate_user(code=code, state=state), loop=self.loop
                ).result(
                    config.auth.callback.timeout
                )
                html = html.replace('SPOTIFY_AUTH_MESSAGE', 'Successfully logged in!')
                html = html.replace('BACKGROUND_COLOR', '#65D46E')
            except Exception as exc:
                logger.exception(exc)
                html = html.replace(
                    'SPOTIFY_AUTH_MESSAGE', 'Could not get authorization token.'
                )
                html = html.replace('BACKGROUND_COLOR', '#EC2E50')
            finally:
                self.callback_reached.set()
            return aiohttp.web.Response(body=html, content_type='text/html')

        web_app.router.add_get('/', callback)
        threading.Thread(target=run_app, daemon=True).start()



#!/usr/bin/env python3
import fire
import kick
from pony.orm import db_session

from .. import APP_NAME, config, logger
from .client import SpotifyClient
from ..mixins.asynch import PlayerMixin, RecommenderMixin

# pylint: disable=too-many-ancestors


class Spotify(SpotifyClient, PlayerMixin, RecommenderMixin):
    """Spotify high-level wrapper."""
    cli = False
    loop = None

    def __init__(self, *args, email=None, username=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.email = email or config.auth.email
        self.username = username or config.auth.username

    def __dir__(self):
        names = super().__dir__()
        names = [name for name in names if not name.startswith('_') and name != 'user']
        return names

    async def auth(self, email=config.auth.email, username=config.auth.username):
        if self.cli and not self.is_authenticated:
            try:
                await self.authenticate(email=email, username=username)
            except Exception as exc:
                logger.exception(exc)
        return self

    @staticmethod
    def update_config(name='config'):
        kick.update_config(APP_NAME.lower(), variant=name)


def main():
    """Main function."""

    import asyncio

    try:
        import uvloop

        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    except ImportError:
        pass
    try:
        Spotify.cli = True
        Spotify.loop = asyncio.get_event_loop()
        spotify = Spotify()
        with db_session:
            fire.Fire(spotify)
    except KeyboardInterrupt:
        print('Quitting')
    finally:
        if spotify.session and not Spotify.loop.is_running():
            Spotify.loop.run_until_complete(spotify.session.close())


if __name__ == '__main__':
    main()



import os
import re
import time
import random
from io import BytesIO
from enum import IntEnum
from uuid import UUID, NAMESPACE_URL, uuid4, uuid5
from datetime import date, datetime
from collections import OrderedDict

import aiohttp
import requests
import psycopg2.extras
from first import first
from pony.orm import (
    Set,
    Json,
    Database,
    Optional,
    Required,
    PrimaryKey,
    ObjectNotFound,
    desc,
    select,
    ormtypes,
    sql_debug,
    db_session,
)
from pycountry import countries
from colorthief import ColorThief
from pony.orm.core import CacheIndexError
from unsplash.errors import UnsplashError
from psycopg2.extensions import register_adapter

from .. import Unsplash, config, logger
from ..constants import TimeRange

register_adapter(ormtypes.TrackedDict, psycopg2.extras.Json)
if os.getenv('SQL_DEBUG'):
    sql_debug(True)
    import logging

    logging.getLogger('pony.orm.sql').setLevel(logging.DEBUG)
db = Database()


class ImageMixin:

    def image(self, width=None, height=None):
        if width:
            image = self.images.select().where(lambda i: i.width >= width).order_by(
                Image.width
            ).first()
        elif height:
            image = self.images.select().where(lambda i: i.height >= height).order_by(
                Image.height
            ).first()
        else:
            image = self.images.select().order_by(desc(Image.width)).first()
        return image

    def get_image_queries(self):
        words = self.name.split()
        stems = [[w[:i] for i in range(len(w), 2, -1)] for w in words]
        queries = [*words, self.name, *sum(stems, [])]
        return [f'{query} music' for query in queries]

    async def fetch_unsplash_image(self, width=None, height=None):
        image = self.image(width, height)
        if image:
            return image

        queries = self.get_image_queries()
        photo = None
        for query in queries:
            try:
                photos = await Unsplash.photo.random(
                    query=query, orientation='squarish'
                )
            except UnsplashError:
                continue

            if photos:
                photo = photos[0]
                break

        else:
            photo = (
                await Unsplash.photo.random(query='music', orientation='squarish')
            )[
                0
            ]
        if photo is None:
            return None

        images = select(i for i in Image if i.unsplash_id == photo.id).for_update()
        if images.exists():
            params = {self.__class__.__name__.lower(): self}
            for image in images:
                image.set(**params)
            return self.image(width, height)

        params = {self.__class__.__name__.lower(): self, 'unsplash_id': photo.id}
        image_exists = False
        for url in (
            photo.urls.full, photo.urls.regular, photo.urls.small, photo.urls.thumb
        ):
            image = Image.get(url=url)
            if image:
                image_exists = True
                image.set(**params)
        if image_exists:
            return self.image(width, height)

        ratio = photo.height / photo.width
        params = {
            self.__class__.__name__.lower(): self,
            'color': photo.color,
            'unsplash_id': photo.id,
            'unsplash_user_fullname': photo.user.name,
            'unsplash_user_username': photo.user.username,
        }
        Image(url=photo.urls.full, width=photo.width, height=photo.height, **params)
        Image(
            url=photo.urls.regular,
            width=Image.REGULAR,
            height=int(round(ratio * Image.REGULAR)),
            **params,
        )
        Image(
            url=photo.urls.small,
            width=Image.SMALL,
            height=int(round(ratio * Image.SMALL)),
            **params,
        )
        Image(
            url=photo.urls.thumb,
            width=Image.THUMB,
            height=int(round(ratio * Image.THUMB)),
            **params,
        )
        return self.image(width, height)


class User(db.Entity, ImageMixin):
    _table_ = 'users'
    DEFAULT_EMAIL = 'spfy@backend'
    DEFAULT_USERNAME = 'spfy-backend'
    DEFAULT_USERID = uuid5(NAMESPACE_URL, DEFAULT_USERNAME)
    id = PrimaryKey(UUID, default=uuid4)  # pylint: disable=redefined-builtin
    email = Optional(str, unique=True, index=True)
    username = Required(str, unique=True, index=True)
    country = Required('Country')
    preferred_country = Optional('Country')
    spotify_user = Optional('SpotifyUser')
    display_name = Optional(str)
    birthdate = Optional(date)
    token = Required(Json, volatile=True)
    api_calls = Required(int, default=0, volatile=True)
    created_at = Required(datetime, default=datetime.now)
    last_usage_at = Required(datetime, default=datetime.now)
    images = Set('Image', cascade_delete=True)
    top_artists = Set('Artist')
    disliked_artists = Set('Artist')
    top_genres = Set('Genre')
    disliked_genres = Set('Genre')
    top_countries = Set('Country')
    disliked_countries = Set('Country')
    top_cities = Set('City')
    disliked_cities = Set('City')
    top_expires_at = Optional(Json, volatile=True)

    @classmethod
    async def from_dict_async(cls, user):
        if user.images:
            try:
                color = await Image.grab_color_async(user.images[-1].url)
            except:
                color = '#000000'
            for image in user.images:
                image.color = color
        return cls.from_dict(user, grab_image_color=False)

    @classmethod
    def from_dict(cls, user, grab_image_color=True):
        if user.images and grab_image_color:
            try:
                color = Image.grab_color(user.images[-1].url)
            except:
                color = '#000000'
            for image in user.images:
                image.color = color
        images = [Image.get(url=image.url) or Image(**image) for image in user.images]
        spotify_user = (
            SpotifyUser.get(id=user.id)
            or SpotifyUser(id=user.id, name=user.get('display_name') or '')
        )
        return cls(
            id=user.user_id,
            spotify_user=spotify_user,
            username=user.id,
            email=user.email,
            token=user.token,
            country=Country.from_str(code=user.country),
            images=images,
            display_name=user.display_name or '',
            birthdate=datetime.strptime(
                user.birthdate, '%Y-%m-%d'
            ) if user.birthdate else None,
        )

    async def _fetch_artist(self, artist, client):
        if artist and client:
            if Artist.exists(id=artist):
                return artist

            artist_result = await client.artist(artist)
            if not artist_result:
                return artist

            try:
                with db_session:
                    # pylint: disable=unused-variable
                    artist_row = await Artist.from_dict_async(artist_result)
            except CacheIndexError:
                pass
        return artist

    async def dislike_async(
        self, artist=None, genre=None, country=None, city=None, client=None
    ):
        artist = await self._fetch_artist(artist, client)
        self.dislike(artist=artist, genre=genre, country=country, city=city)

    async def like_async(
        self, artist=None, genre=None, country=None, city=None, client=None
    ):
        artist = await self._fetch_artist(artist, client)
        self.like(artist=artist, genre=genre, country=country, city=city)

    def dislike(self, artist=None, genre=None, country=None, city=None):
        assert artist or genre or country or city
        if artist:
            artist = Artist.get(id=artist)
            if artist:
                self.disliked_artists.add(artist)
                self.top_artists.remove(artist)
        if genre:
            genre = Genre.get(name=genre) or Genre(name=genre)
            if genre:
                self.disliked_genres.add(genre)
                self.top_genres.remove(genre)
                self.disliked_artists.add(genre.artists)
                self.top_artists.remove(genre.artists)
        if country:
            country = Country.get(code=country)
            if country:
                self.disliked_countries.add(country)
                self.top_countries.remove(country)
        if city:
            city = City.get(name=city)
            if city:
                self.disliked_cities.add(city)
                self.top_cities.remove(city)

    def like(self, artist=None, genre=None, country=None, city=None):
        assert artist or genre or country or city
        if artist:
            artist = Artist.get(id=artist)
            if artist:
                self.top_artists.add(artist)
                self.disliked_artists.remove(artist)
        if genre:
            genre = Genre.get(name=genre)
            if genre:
                self.top_genres.add(genre)
                self.disliked_genres.remove(genre)
                self.top_artists.add(genre.artists)
                self.disliked_artists.remove(genre.artists)
        if country:
            country = Country.get(code=country)
            if country:
                self.top_countries.add(country)
                self.disliked_countries.remove(country)
        if city:
            city = City.get(name=city)
            if city:
                self.top_cities.add(city)
                self.disliked_cities.remove(city)

    def top_expired(self, time_range):
        time_range = TimeRange(time_range).value
        return (
            not self.top_expires_at
            or time_range not in self.top_expires_at
            or time.time() >= self.top_expires_at[time_range]
        )

    @classmethod
    def default(cls):
        try:
            return User[cls.DEFAULT_USERID]

        except ObjectNotFound:
            return User(
                id=cls.DEFAULT_USERID,
                username=cls.DEFAULT_USERNAME,
                email=cls.DEFAULT_EMAIL,
                token={},
                country=Country.from_str(code='US'),
            )

    @staticmethod
    def token_updater(_id):

        @db_session
        def update(token):
            User[_id].token = token

        return update


class Image(db.Entity):
    _table_ = 'images'
    REGULAR = 1080
    SMALL = 400
    THUMB = 200
    url = PrimaryKey(str)
    height = Optional(int)
    width = Optional(int)
    color = Optional(str)
    playlist = Optional('Playlist')
    artist = Optional('Artist')
    genre = Optional('Genre')
    country = Optional('Country')
    city = Optional('City')
    user = Optional('User')
    unsplash_id = Optional(str, index=True)
    unsplash_user_fullname = Optional(str)
    unsplash_user_username = Optional(str)

    # pylint: disable=no-self-use

    def unsplash_url(self):
        return f'https://unsplash.com/?utm_source={config.unsplash.app_name}&utm_medium=referral'

    def unsplash_user_url(self):
        return f'https://unsplash.com/@{self.unsplash_user_username}?utm_source={config.unsplash.app_name}&utm_medium=referral'

    def unsplash_credits(self):
        return {
            'user_name': self.unsplash_user_fullname,
            'user_url': self.unsplash_user_url(),
            'site_url': self.unsplash_url(),
        }

    @staticmethod
    async def grab_color_async(image_url):
        async with aiohttp.ClientSession() as client:
            async with client.get(image_url) as resp:
                image_file = BytesIO(await resp.read())
                color = ColorThief(image_file).get_color(quality=1)
                return f'#{color[0]:02x}{color[1]:02x}{color[2]:02x}'

    @staticmethod
    def grab_color(image_url):
        resp = requests.get(image_url)
        image_file = BytesIO(resp.content)
        color = ColorThief(image_file).get_color(quality=1)
        return f'#{color[0]:02x}{color[1]:02x}{color[2]:02x}'


class SpotifyUser(db.Entity):
    _table_ = 'spotify_users'
    id = PrimaryKey(str)  # pylint: disable=redefined-builtin
    name = Optional(str)
    user = Optional('User')
    playlists = Set('Playlist')

    @property
    def uri(self):
        return f'spotify:user:{self.id}'

    @property
    def href(self):
        return f'https://api.spotify.com/v1/users/{self.id}'

    @property
    def external_url(self):
        return f'http://open.spotify.com/user/{self.id}'


class Genre(db.Entity, ImageMixin):
    _table_ = 'genres'
    name = PrimaryKey(str)
    artists = Set('Artist')
    playlists = Set('Playlist')
    fans = Set(User, reverse='top_genres', table='genre_fans')
    haters = Set(User, reverse='disliked_genres', table='genre_haters')
    images = Set(Image, cascade_delete=True)

    def play(self, client, device=None):
        popularity = random.choice(list(Playlist.Popularity)[:3])
        playlist = client.genre_playlist(self.name, popularity)
        return playlist.play(device=device)


class Country(db.Entity, ImageMixin):
    _table_ = 'countries'
    code = PrimaryKey(str, max_len=2)
    name = Required(str, index=True)
    users = Set('User', reverse='country')
    users_preferring = Set('User', reverse='preferred_country')
    cities = Set('City')
    playlists = Set('Playlist')
    fans = Set(User, reverse='top_countries', table='country_fans')
    haters = Set(User, reverse='disliked_countries', table='country_haters')
    images = Set(Image, cascade_delete=True)

    @staticmethod
    def _name_match(country, name):
        try:
            return name in country.name.lower() or name in country.official_name.lower()

        except:
            return False

    @classmethod
    def from_str(cls, name=None, code=None):
        iso_country = None
        if name == 'UK':
            iso_country = countries.get(alpha_2='GB')
        elif name == 'USA':
            iso_country = countries.get(alpha_2='US')
        elif name is not None:
            try:
                iso_country = countries.get(name=name)
            except KeyError:
                try:
                    iso_country = countries.get(official_name=name)
                except KeyError:
                    iso_country = first(
                        countries, key=lambda c: cls._name_match(c, name.lower())
                    )
        elif code is not None:
            try:
                iso_country = countries.get(alpha_2=code)
            except KeyError:
                iso_country = first(
                    countries, key=lambda c: code.lower() in c.alpha_2.lower()
                )
        if not iso_country:
            logger.error(f'Could not find a country with name={name} and code={code}')
            return None

        return cls.get(name=iso_country.name) or cls(
            name=iso_country.name, code=iso_country.alpha_2
        )

    def get_image_queries(self):
        words = self.name.split()
        stems = [[w[:i] for i in range(len(w), 2, -1)] for w in words]
        queries = [self.name, *words, *sum(stems, [])]
        return queries


class City(db.Entity, ImageMixin):
    _table_ = 'cities'
    name = PrimaryKey(str)
    country = Required(Country)
    playlists = Set('Playlist')
    fans = Set(User, reverse='top_cities', table='city_fans')
    haters = Set(User, reverse='disliked_cities', table='city_haters')
    images = Set(Image, cascade_delete=True)

    def get_image_queries(self):
        words = self.name.split()
        stems = [[w[:i] for i in range(len(w), 2, -1)] for w in words]
        queries = [self.name, self.country.name, *words, *sum(stems, [])]
        return queries


class Playlist(db.Entity, ImageMixin):
    _table_ = 'playlists'
    YEAR = '(?P<year>[0-9]{4})'
    GENRE = '(?P<genre>.+)'
    CITY = '(?P<city>.+)'
    COUNTRY = '(?P<country>.+)'
    COUNTRY_CODE = '(?P<country_code>[A-Z]{2})'
    DATE = '(?P<date>[0-9]{8})'
    INTRO_POPULARITY = '(?P<popularity>Intro)'
    GENRE_POPULARITY_TITLE = '(?P<popularity>Sound|Pulse|Edge)'
    GENRE_POPULARITY_LOWER = '(?P<popularity>sound|pulse|edge)'
    NEEDLE_POPULARITY = '(?P<popularity>Current|Emerging|Underground)'
    PATTERNS = OrderedDict(
        intro_to_genre=re.compile(f'^{INTRO_POPULARITY} to {GENRE}$'),
        sound_of_city=re.compile(f'^The Sound of {CITY} {COUNTRY_CODE}$'),
        needle=re.compile(
            f'^The Needle / {COUNTRY} {DATE}(?: - {NEEDLE_POPULARITY})?$'
        ),
        pine_needle=re.compile(f'^The Pine Needle / {COUNTRY}$'),
        year_in_genre=re.compile(f'^{YEAR} in {GENRE}$'),
        meta_genre=re.compile(f'^Meta{GENRE_POPULARITY_LOWER}: {GENRE}$'),
        meta_year_in_genre=re.compile(f'^Meta{YEAR}: {GENRE}$'),
        sound_of_genre=re.compile(f'^The {GENRE_POPULARITY_TITLE} of {GENRE}$'),
    )

    class Popularity(IntEnum):
        SOUND = 0
        PULSE = 1
        EDGE = 2
        CURRENT = 3
        EMERGING = 4
        UNDERGROUND = 5
        YEAR = 6
        ALL = 7
        INTRO = 8

    id = PrimaryKey(str)  # pylint: disable=redefined-builtin
    collaborative = Required(bool)
    name = Required(str)
    description = Optional(str)
    owner = Required(SpotifyUser)
    public = Required(bool)
    snapshot_id = Required(str)
    tracks = Required(int)
    popularity = Optional(int, index=True)
    genre = Optional(Genre)
    country = Optional(Country)
    city = Optional(City)
    date = Optional(date)
    year = Optional(int)
    christmas = Optional(bool, index=True)
    meta = Optional(bool, index=True)
    images = Set(Image, cascade_delete=True)

    def play(self, client, device=None):
        return client.start_playback(playlist=self.uri, device=device)

    @property
    def uri(self):
        return f'spotify:user:{self.owner.id}:playlist:{self.id}'

    @property
    def href(self):
        return f'https://api.spotify.com/v1/users/{self.owner.id}/playlists/{self.id}'

    @property
    def external_url(self):
        return f'http://open.spotify.com/user/{self.owner.id}/playlist/{self.id}'

    @classmethod
    def get_fields(cls, groups):
        fields = {}
        if 'year' in groups:
            year = groups['year']
            fields['year'] = int(year)
            fields['date'] = datetime(int(year), 1, 1)
            fields['popularity'] = cls.Popularity.YEAR.value
        if 'city' in groups and 'country_code' in groups:
            city = groups['city']
            country_code = groups['country_code']
            fields['country'] = Country.from_str(code=country_code)
            fields['city'] = City.get(name=city) or City(
                name=city, country=fields['country']
            )
            fields['popularity'] = cls.Popularity.SOUND.value
        if 'genre' in groups:
            genre = groups['genre'].lower()
            fields['genre'] = Genre.get(name=genre) or Genre(name=genre)
            if 'christmas' in genre:
                fields['christmas'] = True
        if 'country' in groups:
            country = groups['country']
            fields['country'] = Country.from_str(name=country)
        if 'date' in groups:
            _date = groups['date']
            fields['date'] = datetime.strptime(_date, '%Y%m%d').date()
        if 'popularity' in groups:
            popularity = groups['popularity']
            if popularity:
                fields['popularity'] = cls.Popularity[popularity.upper()].value
            else:
                fields['popularity'] = cls.Popularity.ALL.value
        return fields

    @classmethod
    def from_dict(
        cls, playlist
    ):  # pylint: disable=too-many-return-statements,too-many-statements
        owner = (
            SpotifyUser.get(id=playlist.owner.id)
            or SpotifyUser(
                id=playlist.owner.id, name=playlist.owner.get('display_name') or ''
            )
        )
        fields = {
            'id': playlist.id,
            'collaborative': playlist.collaborative,
            'name': playlist.name,
            'owner': owner,
            'public': playlist.public,
            'snapshot_id': playlist.snapshot_id,
            'tracks': playlist.tracks.total,
            'christmas': 'Pine Needle' in playlist.name
            or 'christmas' in playlist.name.lower(),
            'meta': playlist.name.startswith('Meta'),
            'images': [Image.get(url=im.url) or Image(**im) for im in playlist.images],
        }
        for pattern in cls.PATTERNS.values():
            match = pattern.match(playlist.name)
            if match:
                groups = match.groupdict()
                fields.update(cls.get_fields(groups))
                break

        else:
            logger.warning(f'No pattern matches the playlist: {playlist.name}')
        return cls(**fields)


class Artist(db.Entity, ImageMixin):
    _table_ = 'artists'
    id = PrimaryKey(str)  # pylint: disable=redefined-builtin
    name = Required(str, index=True)
    followers = Required(int)
    genres = Set(Genre)
    fans = Set(User, reverse='top_artists', table='artist_fans')
    haters = Set(User, reverse='disliked_artists', table='artist_haters')
    popularity = Optional(int)
    images = Set(Image, cascade_delete=True)

    def play(self, client, device=None):
        return client.start_playback(artist=self.uri, device=device)

    @property
    def uri(self):
        return f'spotify:artist:{self.id}'

    @property
    def href(self):
        return f'https://api.spotify.com/v1/artists/{self.id}'

    @property
    def external_url(self):
        return f'http://open.spotify.com/artist/{self.id}'

    @classmethod
    async def from_dict_async(cls, artist):
        if artist.images:
            try:
                color = await Image.grab_color_async(artist.images[-1].url)
            except:
                color = '#000000'
            for image in artist.images:
                image.color = color
        return cls.from_dict(artist, grab_image_color=False)

    @classmethod
    def from_dict(cls, artist, grab_image_color=True):
        if artist.images and grab_image_color:
            try:
                color = Image.grab_color(artist.images[-1].url)
            except:
                color = '#000000'
            for image in artist.images:
                image.color = color
        genres = [Genre.get(name=genre) or Genre(name=genre) for genre in artist.genres]
        images = [Image.get(url=image.url) or Image(**image) for image in artist.images]
        return cls(
            id=artist.id,
            name=artist.name,
            followers=artist.followers.total,
            genres=genres,
            images=images,
            popularity=artist.popularity,
        )


# pylint: disable=too-few-public-methods


class AudioFeatures(db.Entity):
    _table_ = 'audio_features'
    KEYS = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
    id = PrimaryKey(str)  # pylint: disable=redefined-builtin
    acousticness = Required(float, min=0.0, max=1.0)
    danceability = Required(float, min=0.0, max=1.0)
    duration_ms = Required(int, min=0)
    energy = Required(float, min=0.0, max=1.0)
    instrumentalness = Required(float, min=0.0, max=1.0)
    key = Required(int, min=0, max=11)
    liveness = Required(float, min=0.0, max=1.0)
    loudness = Required(float, min=-60.0, max=0.0)
    mode = Required(bool)
    speechiness = Required(float, min=0.0, max=1.0)
    tempo = Required(float, min=0, max=1000)
    time_signature = Required(int)
    valence = Required(float, min=0.0, max=1.0)

    def to_dict(self, *args, **kwargs):  # pylint: disable=arguments-differ
        _dict = super().to_dict(*args, **kwargs)
        if 'mode' in _dict:
            _dict['mode'] = int(_dict['mode'])
        return _dict

    @classmethod
    def from_dict(cls, track):
        return cls(
            id=track['id'],
            acousticness=track['acousticness'],
            danceability=track['danceability'],
            duration_ms=track['duration_ms'],
            energy=track['energy'],
            instrumentalness=track['instrumentalness'],
            key=track['key'],
            liveness=track['liveness'],
            loudness=track['loudness'],
            mode=bool(track['mode']),
            speechiness=track['speechiness'],
            tempo=track['tempo'],
            time_signature=track['time_signature'],
            valence=track['valence'],
        )


if config.database.connection.filename:
    config.database.connection.filename = os.path.expandvars(
        config.database.connection.filename
    )
db.bind(**config.database.connection)
if config.database.generate_mapping or os.getenv('SPFY_GENERATE_MAPPING') == 'true':
    db.generate_mapping(create_tables=True)



__version__ = '3.0.16'

import pathlib  # isort:skip

root = pathlib.Path(__file__).parent  # isort:skip
APP_NAME = 'SPFY'

import os  # isort:skip

ENV = os.getenv(f'{APP_NAME.upper()}_ENV', 'config')  # isort:skip

import kick  # isort:skip

kick.start(f'{APP_NAME.lower()}', config_variant=ENV)  # isort:skip

from kick import config, logger  # isort:skip

Unsplash = None
if config.unsplash.auth.client_id:
    from unsplash import Auth, Api

    auth = Auth(**config.unsplash.auth)
    Unsplash = Api(auth)
from .client import SpotifyClient
from .result import SpotifyResult
from .wrapper import Spotify
from .constants import *
from .exceptions import *



# -*- coding: utf-8 -*-
import abc
import asyncio
import subprocess
from time import sleep

from . import logger
from .constants import VOLUME_FADE_SECONDS

try:
    import alsaaudio  # pylint: disable=import-error
except:
    pass


class VolumeControl(abc.ABC):

    @abc.abstractmethod
    def unmute(self):
        pass

    @abc.abstractmethod
    def mute(self):
        pass

    @abc.abstractproperty
    def volume(self):  # pylint: disable=method-hidden
        pass

    @volume.setter
    def volume(self, val):
        pass

    @abc.abstractmethod
    def should_stop_fading(self, device_volume, old_volume):
        return abs(device_volume - old_volume) > 2

    def fade(
        self, limit=100, start=1, step=1, seconds=VOLUME_FADE_SECONDS, force=False
    ):
        try:
            self.unmute()
            delay = seconds / ((limit - start) / step)
            device_volume = self.volume
            self.volume = start
            for next_volume in range(start + step, limit + 1, step):
                sleep(delay)
                device_volume = self.volume
                old_volume = next_volume - step
                if not force and self.should_stop_fading(device_volume, old_volume):
                    logger.debug(
                        f'''A stop fading condition was met:
                        Current volume: {device_volume}
                        Old volume: {old_volume}'''
                    )
                    break

                logger.debug(f'Setting volume to {next_volume}')
                self.volume = next_volume
        except Exception as exc:
            logger.exception(exc)


class SpotifyVolumeControl(VolumeControl):

    def __init__(self, client, device=None):
        self.spotify = client
        self.device = device
        self.old_volume = None

    def mute(self):
        self.old_volume = self.volume
        self.spotify.volume(0, device=self.device)

    def unmute(self):
        if self.old_volume:
            self.volume = self.old_volume
        else:
            self.volume = 1

    def should_stop_fading(self, device_volume, old_volume):
        is_playing = self.spotify.current_playback().is_playing
        logger.debug(f'Spotify playing: {is_playing}')
        return super().should_stop_fading(device_volume, old_volume) or not is_playing

    @property
    def volume(self):
        return int(self.spotify.get_device(device=self.device).volume_percent or 0) + 1

    @volume.setter
    def volume(self, val):
        self.spotify.volume(min(max(val, 1), 100), device=self.device)


class SpotifyVolumeControlAsync(VolumeControl):

    def __init__(self, client, device=None):
        self.spotify = client
        self.device = device
        self.old_volume = None

    async def mute(self):
        self.old_volume = self.volume
        await self.spotify.volume(0, device=self.device)

    async def unmute(self):
        if self.old_volume:
            await self.set_volume(self.old_volume)
        else:
            await self.set_volume(1)

    async def should_stop_fading(self, device_volume, old_volume):
        is_playing = (await self.spotify.current_playback()).is_playing
        logger.debug(f'Spotify playing: {is_playing}')
        return super().should_stop_fading(device_volume, old_volume) or not is_playing

    async def volume(self):  # pylint: disable=method-hidden
        device = await self.spotify.get_device(device=self.device)
        return int(device.volume_percent or 0) + 1

    async def set_volume(self, val):
        await self.spotify.volume(min(max(val, 1), 100), device=self.device)

    async def fade(
        self, limit=100, start=1, step=1, seconds=VOLUME_FADE_SECONDS, force=False
    ):
        try:
            await self.unmute()
            delay = seconds / ((limit - start) / step)
            device_volume = await self.volume()
            await self.set_volume(start)
            for next_volume in range(start + step, limit + 1, step):
                await asyncio.sleep(delay)
                device_volume = await self.volume()
                old_volume = next_volume - step
                should_stop_fading = await self.should_stop_fading(
                    device_volume, old_volume
                )
                if not force and should_stop_fading:
                    logger.debug(
                        f'''A stop fading condition was met:
                        Current volume: {device_volume}
                        Old volume: {old_volume}'''
                    )
                    break

                logger.debug(f'Setting volume to {next_volume}')
                await self.set_volume(next_volume)
        except Exception as exc:
            logger.exception(exc)


class AlsaVolumeControl(VolumeControl):

    def __init__(self, mixer_name, device=None):
        if isinstance(device, str):
            kwargs = dict(device=device)
        elif isinstance(device, int):
            kwargs = dict(cardindex=device)
        else:
            kwargs = dict()
        self.mixer_name = mixer_name
        self.device = device
        self.mixer = alsaaudio.Mixer(mixer_name, **kwargs)

    def mute(self):
        self.mixer.setmute(1)

    def unmute(self):
        self.mixer.setmute(0)

    def should_stop_fading(self, device_volume, old_volume):
        is_mute = self.mixer.getmute()[0]
        logger.debug(f'Mute status: {is_mute}')
        return super().should_stop_fading(device_volume, old_volume) or is_mute

    @property
    def volume(self):
        vol = self.mixer.getvolume()
        return sum(vol) // len(vol)

    @volume.setter
    def volume(self, val):
        self.mixer.setvolume(max(val, 1))


class ApplescriptVolumeControl(VolumeControl):

    def __init__(self, device=None):
        self.device = device
        self.old_volume = None

    def fade(
        self, limit=100, start=1, step=1, seconds=VOLUME_FADE_SECONDS, force=False
    ):
        if self.device:
            self.switch_audio_device(self.device)
        super().fade(limit, start, step, seconds, force)

    @staticmethod
    def osascript(cmd):
        cmd = f"/usr/bin/osascript -e '{cmd}'"
        result = subprocess.check_output(cmd, shell=True).decode().strip()
        logger.debug('Command: %s', cmd)
        logger.debug('Output: %s', result)
        return result

    @staticmethod
    def switch_audio_device(device):
        logger.debug(f'Switching audio device: {device}')
        subprocess.call(f"/usr/local/bin/SwitchAudioSource -s \"{device}\"", shell=True)

    def should_stop_fading(self, device_volume, old_volume):
        is_playing = self.osascript(
            'tell application "Spotify" to get player state'
        ) == 'playing'
        logger.debug(f'Spotify playing: {is_playing}')
        return super().should_stop_fading(device_volume, old_volume) or not is_playing

    def spotify_mute(self):
        self.old_volume = self.spotify_volume
        self.osascript('tell application "Spotify" to set sound volume to 0')

    def spotify_unmute(self):
        if self.old_volume:
            self.spotify_volume = self.old_volume
        else:
            self.spotify_volume = 1

    def mute(self):
        self.osascript('set volume with output muted')

    def unmute(self):
        self.osascript('set volume without output muted')

    @property
    def spotify_volume(self):
        return int(self.osascript('tell application "Spotify" to get sound volume')) + 1

    @spotify_volume.setter
    def spotify_volume(self, val):
        self.osascript(
            f'tell application "Spotify" to set sound volume to {max(val, 1)}'
        )

    @property
    def system_volume(self):
        return int(self.osascript('output volume of (get volume settings)'))

    @system_volume.setter
    def system_volume(self, val):
        self.osascript(f'set volume output volume {max(val, 1)}')

    @property
    def volume(self):
        return self.system_volume

    @volume.setter
    def volume(self, val):
        self.old_volume = self.volume
        self.system_volume = val
        self.spotify_volume = val // 2


class LinuxVolumeControl(AlsaVolumeControl):

    def __init__(
        self, spotify_client, alsa_mixer_name, spotify_device=None, alsa_device=None
    ):
        super().__init__(alsa_mixer_name, device=alsa_device)
        self.spotify_volume_control = SpotifyVolumeControl(
            spotify_client, spotify_device
        )

    def should_stop_fading(self, device_volume, old_volume):
        return (
            super().should_stop_fading(device_volume, old_volume)
            or self.spotify_should_stop_fading(device_volume, old_volume)
        )

    def spotify_should_stop_fading(self, device_volume, old_volume):
        return self.spotify_volume_control.should_stop_fading(device_volume, old_volume)


class LinuxVolumeControlAsync(AlsaVolumeControl):

    def __init__(
        self, spotify_client, alsa_mixer_name, spotify_device=None, alsa_device=None
    ):
        super().__init__(alsa_mixer_name, device=alsa_device)
        self.spotify_volume_control = SpotifyVolumeControlAsync(
            spotify_client, spotify_device
        )

    async def should_stop_fading(self, device_volume, old_volume):
        return (
            super().should_stop_fading(device_volume, old_volume)
            or await self.spotify_should_stop_fading(device_volume, old_volume)
        )

    async def spotify_should_stop_fading(self, device_volume, old_volume):
        return await self.spotify_volume_control.should_stop_fading(
            device_volume, old_volume
        )

    async def fade(
        self, limit=100, start=1, step=1, seconds=VOLUME_FADE_SECONDS, force=False
    ):
        try:
            self.unmute()
            delay = seconds / ((limit - start) / step)
            device_volume = self.volume
            self.volume = start
            for next_volume in range(start + step, limit + 1, step):
                sleep(delay)
                device_volume = self.volume
                old_volume = next_volume - step
                should_stop_fading = await self.should_stop_fading(
                    device_volume, old_volume
                )
                if not force and should_stop_fading:
                    logger.debug(
                        f'''A stop fading condition was met:
                        Current volume: {device_volume}
                        Old volume: {old_volume}'''
                    )
                    break

                logger.debug(f'Setting volume to {next_volume}')
                self.volume = next_volume
        except Exception as exc:
            logger.exception(exc)



import re
from enum import Enum


class ItemType(Enum):
    TRACKS = 'tracks'
    TRACK = 'track'
    PLAYLIST = 'playlist'
    ARTIST = 'artist'
    ALBUM = 'album'


class TimeRange(Enum):
    SHORT_TERM = 'short_term'
    MEDIUM_TERM = 'medium_term'
    LONG_TERM = 'long_term'


class VolumeBackend(Enum):
    SPOTIFY = 'spotify'
    APPLESCRIPT = 'applescript'
    ALSA = 'alsa'
    LINUX = 'linux'


class AuthFlow(Enum):
    AUTHORIZATION_CODE = 'authorization_code'
    CLIENT_CREDENTIALS = 'client_credentials'


class AudioFeature(Enum):
    ACOUSTICNESS = 'acousticness'
    DANCEABILITY = 'danceability'
    DURATION_MS = 'duration_ms'
    ENERGY = 'energy'
    INSTRUMENTALNESS = 'instrumentalness'
    KEY = 'key'
    LIVENESS = 'liveness'
    LOUDNESS = 'loudness'
    MODE = 'mode'
    POPULARITY = 'popularity'
    SPEECHINESS = 'speechiness'
    TEMPO = 'tempo'
    TIME_SIGNATURE = 'time_signature'
    VALENCE = 'valence'


class Scope(Enum):
    PLAYLIST_READ_PRIVATE = 'playlist-read-private'
    PLAYLIST_READ_COLLABORATIVE = 'playlist-read-collaborative'
    PLAYLIST_MODIFY_PUBLIC = 'playlist-modify-public'
    PLAYLIST_MODIFY_PRIVATE = 'playlist-modify-private'
    STREAMING = 'streaming'
    UGC_IMAGE_UPLOAD = 'ugc-image-upload'
    USER_FOLLOW_MODIFY = 'user-follow-modify'
    USER_FOLLOW_READ = 'user-follow-read'
    USER_LIBRARY_READ = 'user-library-read'
    USER_LIBRARY_MODIFY = 'user-library-modify'
    USER_READ_PRIVATE = 'user-read-private'
    USER_READ_BIRTHDATE = 'user-read-birthdate'
    USER_READ_EMAIL = 'user-read-email'
    USER_TOP_READ = 'user-top-read'
    USER_READ_PLAYBACK_STATE = 'user-read-playback-state'
    USER_MODIFY_PLAYBACK_STATE = 'user-modify-playback-state'
    USER_READ_CURRENTLY_PLAYING = 'user-read-currently-playing'
    USER_READ_RECENTLY_PLAYED = 'user-read-recently-played'


AllScopes = [scope.value for scope in Scope]


class API(Enum):
    PREFIX = 'https://api.spotify.com'
    ALBUM = '/v1/albums/{id}'
    ALBUM_TRACKS = '/v1/albums/{id}/tracks'
    ALBUMS = '/v1/albums'
    ARTIST = '/v1/artists/{id}'
    ARTIST_ALBUMS = '/v1/artists/{id}/albums'
    ARTIST_RELATED_ARTISTS = '/v1/artists/{id}/related-artists'
    ARTIST_TOP_TRACKS = '/v1/artists/{id}/top-tracks'
    ARTISTS = '/v1/artists'
    AUDIO_ANALYSIS = '/v1/audio-analysis/{id}'
    AUDIO_FEATURES_SINGLE = '/v1/audio-features/{id}'
    AUDIO_FEATURES_MULTIPLE = '/v1/audio-features'
    RECOMMENDATIONS = '/v1/recommendations'
    RECOMMENDATIONS_GENRES = '/v1/recommendations/available-genre-seeds'
    AUTHORIZE = 'https://accounts.spotify.com/authorize'
    TOKEN = 'https://accounts.spotify.com/api/token'
    CATEGORIES = '/v1/browse/categories'
    CATEGORY = '/v1/browse/categories/{id}'
    CATEGORY_PLAYLISTS = '/v1/browse/categories/{id}/playlists'
    FEATURED_PLAYLISTS = '/v1/browse/featured-playlists'
    NEW_RELEASES = '/v1/browse/new-releases'
    ME = '/v1/me'
    MY_ALBUMS = '/v1/me/albums'
    MY_ALBUMS_CONTAINS = '/v1/me/albums/contains'
    MY_FOLLOWING = '/v1/me/following'
    MY_FOLLOWING_CONTAINS = '/v1/me/following/contains'
    MY_PLAYLISTS = '/v1/me/playlists'
    MY_TOP = '/v1/me/top/{type}'
    MY_TRACKS = '/v1/me/tracks'
    MY_TRACKS_CONTAINS = '/v1/me/tracks/contains'
    PLAYLIST = '/v1/users/{user_id}/playlists/{playlist_id}'
    PLAYLISTS = '/v1/users/{user_id}/playlists'
    PLAYLIST_FOLLOWERS = '/v1/users/{owner_id}/playlists/{playlist_id}/followers'
    PLAYLIST_FOLLOWERS_CONTAINS = '/v1/users/{user_id}/playlists/{playlist_id}/followers/contains'
    PLAYLIST_IMAGES = '/v1/users/{user_id}/playlists/{playlist_id}/images'
    PLAYLIST_TRACKS = '/v1/users/{user_id}/playlists/{playlist_id}/tracks'
    SEARCH_ALBUM = '/v1/search?type=album'
    SEARCH_ARTIST = '/v1/search?type=artist'
    SEARCH_PLAYLIST = '/v1/search?type=playlist'
    SEARCH_TRACK = '/v1/search?type=track'
    TRACK = '/v1/tracks/{id}'
    TRACKS = '/v1/tracks'
    USER = '/v1/users/{user_id}'
    CURRENTLY_PLAYING = '/v1/me/player/currently-playing'
    DEVICES = '/v1/me/player/devices'
    NEXT = '/v1/me/player/next'
    PAUSE = '/v1/me/player/pause'
    PLAY = '/v1/me/player/play'
    PLAYER = '/v1/me/player'
    PREVIOUS = '/v1/me/player/previous'
    RECENTLY_PLAYED = '/v1/me/player/recently-played'
    REPEAT = '/v1/me/player/repeat'
    SEEK = '/v1/me/player/seek'
    SHUFFLE = '/v1/me/player/shuffle'
    VOLUME = '/v1/me/player/volume'


VOLUME_FADE_SECONDS = 5 * 60
DEVICE_ID_RE = re.compile(r'[a-zA-Z0-9]{40}')
PLAYLIST_URI_RE = re.compile(r'spotify:user:[^:]+:playlist:[^:]+')
MAX_KEY = 11
MAX_LOUDNESS = -60.0
MAX_POPULARITY = 100



import logging

import aiohttp
from oauthlib.common import urldecode, generate_token

from oauthlib.oauth2 import (  # isort:skip
    TokenExpiredError, WebApplicationClient, InsecureTransportError, is_secure_transport
)

log = logging.getLogger(__name__)


class TokenUpdated(Warning):

    def __init__(self, token):
        super(TokenUpdated, self).__init__()
        self.token = token


class OAuth2Session(aiohttp.ClientSession):
    """Versatile OAuth 2 extension to :class:`aiohttp.ClientSession`.

    Supports any grant type adhering to :class:`oauthlib.oauth2.Client` spec
    including the four core OAuth 2 grants.

    Can be used to create authorization urls, fetch tokens and access protected
    resources using the :class:`aiohttp.ClientSession` interface you are used to.

    - :class:`oauthlib.oauth2.WebApplicationClient` (default): Authorization Code Grant
    - :class:`oauthlib.oauth2.MobileApplicationClient`: Implicit Grant
    - :class:`oauthlib.oauth2.LegacyApplicationClient`: Password Credentials Grant
    - :class:`oauthlib.oauth2.BackendApplicationClient`: Client Credentials Grant

    Note that the only time you will be using Implicit Grant from python is if
    you are driving a user agent able to obtain URL fragments.
    """

    #  pylint: disable=too-many-arguments

    def __init__(
        self,
        client_id=None,
        client=None,
        auto_refresh_url=None,
        auto_refresh_kwargs=None,
        scope=None,
        redirect_uri=None,
        token=None,
        state=None,
        token_updater=None,
        **kwargs,
    ):
        """Construct a new OAuth 2 client session.

        :param client_id: Client id obtained during registration
        :param client: :class:`oauthlib.oauth2.Client` to be used. Default is
                       WebApplicationClient which is useful for any
                       hosted application but not mobile or desktop.
        :param scope: List of scopes you wish to request access to
        :param redirect_uri: Redirect URI you registered as callback
        :param token: Token dictionary, must include access_token
                      and token_type.
        :param state: State string used to prevent CSRF. This will be given
                      when creating the authorization url and must be supplied
                      when parsing the authorization response.
                      Can be either a string or a no argument callable.
        :auto_refresh_url: Refresh token endpoint URL, must be HTTPS. Supply
                           this if you wish the client to automatically refresh
                           your access tokens.
        :auto_refresh_kwargs: Extra arguments to pass to the refresh token
                              endpoint.
        :token_updater: Method with one argument, token, to be used to update
                        your token database on automatic token refresh. If not
                        set a TokenUpdated warning will be raised when a token
                        has been refreshed. This warning will carry the token
                        in its token argument.
        :param kwargs: Arguments to pass to the Session constructor.
        """
        super(OAuth2Session, self).__init__(**kwargs)
        self._client = client or WebApplicationClient(client_id, token=token)
        self.token = token or {}
        self.scope = scope
        self.redirect_uri = redirect_uri
        self.state = state or generate_token
        self._state = state
        self.auto_refresh_url = auto_refresh_url
        self.auto_refresh_kwargs = auto_refresh_kwargs or {}
        self.token_updater = token_updater
        # Allow customizations for non compliant providers through various
        # hooks to adjust requests and responses.
        self.compliance_hook = {
            'access_token_response': set(),
            'refresh_token_response': set(),
            'protected_request': set(),
        }

    def new_state(self):
        """Generates a state string to be used in authorizations."""
        try:
            self._state = self.state()
            log.debug(f'Generated new state {self._state}.')
        except TypeError:
            self._state = self.state
            log.debug(f'Re-using previously supplied state {self._state}.')
        return self._state

    @property
    def client_id(self):
        return getattr(self._client, "client_id", None)

    @client_id.setter
    def client_id(self, value):
        self._client.client_id = value

    @client_id.deleter
    def client_id(self):
        del self._client.client_id

    @property
    def token(self):
        return getattr(self._client, "token", None)

    @token.setter
    def token(self, value):
        self._client.token = value
        self._client._populate_attributes(value)

    @property
    def access_token(self):
        return getattr(self._client, "access_token", None)

    @access_token.setter
    def access_token(self, value):
        self._client.access_token = value

    @access_token.deleter
    def access_token(self):
        del self._client.access_token

    @property
    def authorized(self):
        """Boolean that indicates whether this session has an OAuth token
        or not. If `self.authorized` is True, you can reasonably expect
        OAuth-protected requests to the resource to succeed. If
        `self.authorized` is False, you need the user to go through the OAuth
        authentication dance before OAuth-protected requests to the resource
        will succeed.
        """
        return bool(self.access_token)

    def authorization_url(self, url, state=None, **kwargs):
        """Form an authorization URL.

        :param url: Authorization endpoint url, must be HTTPS.
        :param state: An optional state string for CSRF protection. If not
                      given it will be generated for you.
        :param kwargs: Extra parameters to include.
        :return: authorization_url, state
        """
        state = state or self.new_state()
        return self._client.prepare_request_uri(
            url, redirect_uri=self.redirect_uri, scope=self.scope, state=state, **kwargs
        ), state

    #  pylint: disable=too-many-locals

    async def fetch_token(
        self,
        token_url,
        code=None,
        authorization_response=None,
        body='',
        auth=None,
        username=None,
        password=None,
        method='POST',
        timeout=None,
        headers=None,
        verify_ssl=True,
        proxy=None,
        **kwargs,
    ):
        """Generic method for fetching an access token from the token endpoint.

        If you are using the MobileApplicationClient you will want to use
        token_from_fragment instead of fetch_token.

        :param token_url: Token endpoint URL, must use HTTPS.
        :param code: Authorization code (used by WebApplicationClients).
        :param authorization_response: Authorization response URL, the callback
                                       URL of the request back to you. Used by
                                       WebApplicationClients instead of code.
        :param body: Optional application/x-www-form-urlencoded body to add the
                     include in the token request. Prefer kwargs over body.
        :param auth: An auth tuple or method as accepted by aiohttp.
        :param username: Username used by LegacyApplicationClients.
        :param password: Password used by LegacyApplicationClients.
        :param method: The HTTP method used to make the request. Defaults
                       to POST, but may also be GET. Other methods should
                       be added as needed.
        :param headers: Dict to default request headers with.
        :param timeout: Timeout of the request in seconds.
        :param verify_ssl: Verify SSL certificate.
        :param kwargs: Extra parameters to include in the token request.
        :return: A token dict
        """
        if not is_secure_transport(token_url):
            raise InsecureTransportError()

        if not code and authorization_response:
            self._client.parse_request_uri_response(
                authorization_response, state=self._state
            )
            code = self._client.code
        elif not code and isinstance(self._client, WebApplicationClient):
            code = self._client.code
            if not code:
                raise ValueError(
                    'Please supply either code or ' 'authorization_response parameters.'
                )

        body = self._client.prepare_request_body(
            code=code,
            body=body,
            redirect_uri=self.redirect_uri,
            username=username,
            password=password,
            **kwargs,
        )
        client_id = kwargs.get('client_id', '')
        if auth is None:
            if client_id:
                log.debug(
                    f'Encoding client_id "{client_id}" with client_secret as Basic auth credentials.'
                )
                client_secret = kwargs.get('client_secret', '')
                client_secret = client_secret if client_secret is not None else ''
                auth = aiohttp.BasicAuth(client_id, client_secret)
            elif username:
                if password is None:
                    raise ValueError('Username was supplied, but not password.')

                log.debug('Encoding username, password as Basic auth credentials.')
                auth = aiohttp.BasicAuth(username, password)
        headers = headers or {
            'Accept': 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8',
        }
        self.token = {}
        if method.upper() == 'POST':
            req = self.post(
                token_url,
                data=dict(urldecode(body)),
                timeout=timeout,
                headers=headers,
                auth=auth,
                verify_ssl=verify_ssl,
                proxy=proxy,
            )
            log.debug(f'Prepared fetch token request body {body}')
        elif method.upper() == 'GET':
            # if method is not 'POST', switch body to querystring and GET
            req = self.get(
                token_url,
                params=dict(urldecode(body)),
                timeout=timeout,
                headers=headers,
                auth=auth,
                verify_ssl=verify_ssl,
                proxy=proxy,
            )
            log.debug(f'Prepared fetch token request querystring {body}')
        else:
            raise ValueError('The method kwarg must be POST or GET.')

        async with req as resp:
            text = await resp.text()
            log.debug(f'Request to fetch token completed with status {resp.status}.')
            log.debug(f'Request headers were {resp.request_info.headers}')
            log.debug(f'Response headers were {resp.headers} and content {text}.')
            log.debug(
                f'Invoking {len(self.compliance_hook["access_token_response"])} token response hooks.'
            )
            for hook in self.compliance_hook['access_token_response']:
                log.debug(f'Invoking hook {hook}.')
                resp = hook(resp)
            self._client.parse_request_body_response(text, scope=self.scope)
            self.token = self._client.token
            log.debug(f'Obtained token {self.token}.')
        return self.token

    def token_from_fragment(self, authorization_response):
        """Parse token from the URI fragment, used by MobileApplicationClients.

        :param authorization_response: The full URL of the redirect back to you
        :return: A token dict
        """
        self._client.parse_request_uri_response(
            authorization_response, state=self._state
        )
        self.token = self._client.token
        return self.token

    async def refresh_token(
        self,
        token_url,
        refresh_token=None,
        body='',
        auth=None,
        timeout=None,
        headers=None,
        verify_ssl=True,
        proxy=None,
        **kwargs,
    ):
        """Fetch a new access token using a refresh token.

        :param token_url: The token endpoint, must be HTTPS.
        :param refresh_token: The refresh_token to use.
        :param body: Optional application/x-www-form-urlencoded body to add the
                     include in the token request. Prefer kwargs over body.
        :param auth: An auth tuple or method as accepted by aiohttp.
        :param timeout: Timeout of the request in seconds.
        :param verify_ssl: Verify SSL certificate.
        :param kwargs: Extra parameters to include in the token request.
        :return: A token dict
        """
        if not token_url:
            raise ValueError('No token endpoint set for auto_refresh.')

        if not is_secure_transport(token_url):
            raise InsecureTransportError()

        refresh_token = refresh_token or self.token.get('refresh_token')
        log.debug(f'Adding auto refresh key word arguments {self.auto_refresh_kwargs}.')
        kwargs.update(self.auto_refresh_kwargs)
        body = self._client.prepare_refresh_body(
            body=body, refresh_token=refresh_token, scope=self.scope, **kwargs
        )
        log.debug(f'Prepared refresh token request body {body}')
        if headers is None:
            headers = {
                'Accept': 'application/json',
                'Content-Type': ('application/x-www-form-urlencoded;charset=UTF-8'),
            }
        async with self.post(
            token_url,
            data=dict(urldecode(body)),
            auth=auth,
            timeout=timeout,
            headers=headers,
            verify_ssl=verify_ssl,
            withhold_token=True,
            proxy=proxy,
        ) as resp:
            text = await resp.text()
            log.debug(f'Request to refresh token completed with status {resp.status}.')
            log.debug(f'Response headers were {resp.headers} and content {text}.')
            log.debug(
                f'Invoking {len(self.compliance_hook["refresh_token_response"])} token response hooks.'
            )
            for hook in self.compliance_hook['refresh_token_response']:
                log.debug(f'Invoking hook {hook}.')
                resp = hook(resp)
            self.token = self._client.parse_request_body_response(
                text, scope=self.scope
            )
            if 'refresh_token' not in self.token:
                log.debug('No new refresh token given. Re-using old.')
                self.token['refresh_token'] = refresh_token
        return self.token

    #  pylint: disable=arguments-differ

    async def _request(
        self,
        method,
        url,
        data=None,
        headers=None,
        withhold_token=False,
        client_id=None,
        client_secret=None,
        **kwargs,
    ):
        """Intercept all requests and add the OAuth 2 token if present."""
        if not is_secure_transport(url):
            raise InsecureTransportError()

        if self.token and not withhold_token:
            log.debug(
                'Invoking %d protected resource request hooks.',
                len(self.compliance_hook['protected_request']),
            )
            for hook in self.compliance_hook['protected_request']:
                log.debug(f'Invoking hook {hook}.')
                url, headers, data = hook(url, headers, data)
            log.debug(f'Adding token {self.token} to request.')
            try:
                url, headers, data = self._client.add_token(
                    url, http_method=method, body=data, headers=headers
                )
            # Attempt to retrieve and save new access token if expired
            except TokenExpiredError:
                if self.auto_refresh_url:
                    log.debug(
                        f'Auto refresh is set, attempting to refresh at {self.auto_refresh_url}.'
                    )
                    # We mustn't pass auth twice.
                    auth = kwargs.pop('auth', None)
                    if client_id and client_secret and (auth is None):
                        log.debug(
                            f'Encoding client_id "{client_id}" with client_secret as Basic auth credentials.'
                        )
                        auth = aiohttp.BasicAuth(client_id, client_secret)
                    token = await self.refresh_token(
                        self.auto_refresh_url, auth=auth, **kwargs
                    )
                    if self.token_updater:
                        log.debug(
                            f'Updating token to {token} using {self.token_updater}.'
                        )
                        self.token_updater(token)
                        url, headers, data = self._client.add_token(
                            url, http_method=method, body=data, headers=headers
                        )
                    else:
                        raise TokenUpdated(token)

                else:
                    raise

        log.debug(f'Requesting url {url} using method {method}.')
        log.debug(f'Supplying headers {headers} and data {data}')
        log.debug(f'Passing through key word arguments {kwargs}.')
        return await super(OAuth2Session, self)._request(
            method, url, headers=headers, data=data, **kwargs
        )

    def register_compliance_hook(self, hook_type, hook):
        """Register a hook for request/response tweaking.

        Available hooks are:
            access_token_response invoked before token parsing.
            refresh_token_response invoked before refresh token parsing.
            protected_request invoked before making a request.

        If you find a new hook is needed please send a GitHub PR request
        or open an issue.
        """
        if hook_type not in self.compliance_hook:
            raise ValueError(f'Hook type {hook_type} is not in {self.compliance_hook}.')

        self.compliance_hook[hook_type].add(hook)
